/*
 * Implement AES algorithm in Intel AES-NI instructions.
 *
 * The white paper of AES-NI instructions can be downloaded from:
 *   http://softwarecommunity.intel.com/isn/downloads/intelavx/AES-Instructions-Set_WP.pdf
 *
 * Copyright (C) 2008, Intel Corp.
 *    Author: Huang Ying <ying.huang@intel.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

.text

#include <linux/linkage.h>

key_expansion_128:
	movaps %xmm1, %xmm4
	psrldq $12, %xmm1
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm1, %xmm0

	movaps %xmm0, (%rcx)
	add $0x10, %rcx
	ret

key_expansion_192:
	pshufd $0b01010101, %xmm1, %xmm1
	movaps %xmm1, %xmm4
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm1, %xmm0

	pshufd $0b11111111, %xmm0, %xmm3
	pxor %xmm2, %xmm3
	palignr $12, %xmm0, %xmm3
	pxor %xmm2, %xmm3

	test %r9, %r9
	not %r9
	jnz 1f

	movaps %xmm0, %xmm1
	pslldq $8, %xmm2
	palignr $8, %xmm2, %xmm1
	movaps %xmm1, (%rcx)
	add $0x10, %rcx
	movaps %xmm3, %xmm2
	palignr $8, %xmm0, %xmm3
	movaps %xmm3, (%rcx)
	add $0x10, %rcx
	ret
1:
	movaps %xmm0, (%rcx)
	add $0x10, %rcx
	movaps %xmm3, %xmm2
	ret

key_expansion_256:
	movaps %xmm1, %xmm4
	psrldq $12, %xmm1
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm0, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm1, %xmm0

	movaps %xmm0, (%rcx)
	add $0x10, %rcx

	test %r9, %r9
	jnz 1f

	# aeskeygenassist $0x1, %xmm0, %xmm1
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x01

	pshufd $0b10101010, %xmm1, %xmm1
	movaps %xmm1, %xmm4
	pxor %xmm2, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm2, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm2, %xmm1
	palignr $12, %xmm4, %xmm1
	pxor %xmm1, %xmm2

	movaps %xmm2, (%rcx)
	add $0x10, %rcx
1:
	ret

ENTRY(intel_aes_set_key)
	movups (%rsi), %xmm0		# user key (first 16 bytes)
	movaps %xmm0, (%rdi)
	lea 0x10(%rdi), %rcx		# key addr
	mov %edx, 480(%rdi)		# key len
	cmp $24, %dl
	jb 2f
	je 1f
	movups 0x10(%rsi), %xmm2		# other user key
	movaps %xmm2, (%rcx)
	lea 0x10(%rcx), %rcx
	xor %r9, %r9
	# aeskeygenassist $0x1, %xmm2, %xmm1	# round 1
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x01
	call key_expansion_256
	# aeskeygenassist $0x2, %xmm2, %xmm1	# round 2
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x02
	call key_expansion_256
	# aeskeygenassist $0x4, %xmm2, %xmm1	# round 3
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x04
	call key_expansion_256
	# aeskeygenassist $0x8, %xmm2, %xmm1	# round 4
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x08
	call key_expansion_256
	# aeskeygenassist $0x10, %xmm2, %xmm1	# round 5
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x10
	call key_expansion_256
	# aeskeygenassist $0x20, %xmm2, %xmm1	# round 6
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x20
	call key_expansion_256
	# aeskeygenassist $0x40, %xmm2, %xmm1	# round 7
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x40
	not %r9
	call key_expansion_256
	lea 224(%rdi), %rdx
	jmp 3f
1:
	mov $192, %edx
	movq 0x10(%rsi), %xmm2			# other user key
	xor %r9, %r9
	# aeskeygenassist $0x1, %xmm2, %xmm1	# round 1
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x01
	call key_expansion_192
	# aeskeygenassist $0x2, %xmm2, %xmm1	# round 2
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x02
	call key_expansion_192
	# aeskeygenassist $0x4, %xmm2, %xmm1	# round 3
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x04
	call key_expansion_192
	# aeskeygenassist $0x8, %xmm2, %xmm1	# round 4
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x08
	call key_expansion_192
	# aeskeygenassist $0x10, %xmm2, %xmm1	# round 5
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x10
	call key_expansion_192
	# aeskeygenassist $0x20, %xmm2, %xmm1	# round 6
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x20
	call key_expansion_192
	# aeskeygenassist $0x40, %xmm2, %xmm1	# round 7
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x40
	call key_expansion_192
	# aeskeygenassist $0x80, %xmm2, %xmm1	# round 8
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xca, 0x80
	call key_expansion_192
	lea 192(%rdi), %rdx
	jmp 3f
2:
	# aeskeygenassist $0x1, %xmm0, %xmm1	# round 1
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x01
	call key_expansion_128
	# aeskeygenassist $0x2, %xmm0, %xmm1	# round 2
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x02
	call key_expansion_128
	# aeskeygenassist $0x4, %xmm0, %xmm1	# round 3
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x04
	call key_expansion_128
	# aeskeygenassist $0x8, %xmm0, %xmm1	# round 4
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x08
	call key_expansion_128
	# aeskeygenassist $0x10, %xmm0, %xmm1	# round 5
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x10
	call key_expansion_128
	# aeskeygenassist $0x20, %xmm0, %xmm1	# round 6
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x20
	call key_expansion_128
	# aeskeygenassist $0x40, %xmm0, %xmm1	# round 7
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x40
	call key_expansion_128
	# aeskeygenassist $0x80, %xmm0, %xmm1	# round 8
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x80
	call key_expansion_128
	# aeskeygenassist $0x1b, %xmm0, %xmm1	# round 9
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x1b
	call key_expansion_128
	# aeskeygenassist $0x36, %xmm0, %xmm1	# round 10
	.byte 0x66, 0x0f, 0x3a, 0xdf, 0xc8, 0x36
	call key_expansion_128
	lea 160(%rdi), %rdx
3:
	movaps (%rdi), %xmm0
	movaps (%rdx), %xmm1
	movaps %xmm0, 240(%rdx)
	movaps %xmm1, 240(%rdi)
	lea 0x10(%rdi), %rdi
	lea 224(%rdx), %rsi
4:
	movaps (%rdi), %xmm0
	# aesimc %xmm0, %xmm1
	.byte 0x66, 0x0f, 0x38, 0xdb, 0xc8
	movaps %xmm1, (%rsi)
	lea 0x10(%rdi), %rdi
	lea -0x10(%rsi), %rsi
	cmp %rdx, %rdi
	jb 4b
	xor %rax, %rax
	ret
END(intel_aes_set_key)

ENTRY(intel_aes_enc)
	movups (%rdx), %xmm0		# input
	mov 480(%rdi), %ecx		# key length
	movaps (%rdi), %xmm1		# key
	pxor %xmm1, %xmm0		# round 0
	lea 0x30(%rdi), %rdi
	cmp $24, %cl
	jb 2f
	lea 0x20(%rdi), %rdi
	je 1f
	lea 0x20(%rdi), %rdi
	movaps -0x60(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps -0x50(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
1:
	movaps -0x40(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps -0x30(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
2:
	movaps -0x20(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps -0x10(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps (%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps 0x10(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps 0x20(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps 0x30(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps 0x40(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps 0x50(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps 0x60(%rdi), %xmm1
	# aesenc %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xdc, 0xc1
	movaps 0x70(%rdi), %xmm1
	# aesenclast %xmm1, %xmm0	# last round
	.byte 0x66, 0x0f, 0x38, 0xdd, 0xc1
	movups %xmm0, (%rsi)		# output
	ret
END(intel_aes_enc)

ENTRY(intel_aes_dec)
	movups (%rdx), %xmm0		# input
	mov 480(%rdi), %ecx		# key length
	lea 240(%rdi), %rdi
	movaps (%rdi), %xmm1		# key
	pxor %xmm1, %xmm0		# round 0
	lea 0x30(%rdi), %rdi
	cmp $24, %cl
	jb 2f
	lea 0x20(%rdi), %rdi
	je 1f
	lea 0x20(%rdi), %rdi
	movaps -0x60(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps -0x50(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
1:
	movaps -0x40(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps -0x30(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
2:
	movaps -0x20(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps -0x10(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps (%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps 0x10(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps 0x20(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps 0x30(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps 0x40(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps 0x50(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps 0x60(%rdi), %xmm1
	# aesdec %xmm1, %xmm0
	.byte 0x66, 0x0f, 0x38, 0xde, 0xc1
	movaps 0x70(%rdi), %xmm1
	# aesdeclast %xmm1, %xmm0	# last round
	.byte 0x66, 0x0f, 0x38, 0xdf, 0xc1
	movups %xmm0, (%rsi)		# output
	ret
END(intel_aes_dec)
