#ifndef _AFXXXC_H
#define _AFXXXC_H

/* AFC IOCTL Structures */
#define	AFC_MAGIC_NUMBER	't'
#define	AFC_MINNR		10
#define	AFC_MAXNR		13

#define	AFC_NMI0		_IO(AFC_MAGIC_NUMBER,0)
#define	AFC_NMI1		_IO(AFC_MAGIC_NUMBER,1)
#define	AFC_NMI2		_IO(AFC_MAGIC_NUMBER,2)
#define	AFC_NMI3		_IO(AFC_MAGIC_NUMBER,3)

#define	AFC_SCSI_SEND_COMMAND_0 _IOWR(AFC_MAGIC_NUMBER,10,unsigned long)
#define	AFC_SCSI_SEND_COMMAND_1 _IOWR(AFC_MAGIC_NUMBER,11,unsigned long)
#define	AFC_SCSI_SEND_COMMAND_2 _IOWR(AFC_MAGIC_NUMBER,12,unsigned long)
#define	AFC_SCSI_SEND_COMMAND_3 _IOWR(AFC_MAGIC_NUMBER,13,unsigned long)

#define	AFC_MINOR		255

#define	AFC_IOCTL_DONE		1

struct afc_ioctl_command {
	unsigned short	inlen;
	unsigned char	target;
	unsigned char	channel;
	unsigned short	outlen;
	unsigned char	rsv0;
	unsigned char	rsv1;
	unsigned char	data[0];
};

#ifdef __KERNEL__

#if !defined(LINUX_VERSION_CODE)
#include <linux/version.h>
#endif

/* Tunable parameters */

/* How many controllers */
#define MAX_AFC_HOSTS	 	4

/* How many channels */
#define MAX_AFC_CHANNELS	2
#define AFC_LOGICAL_CHANNEL	0

/* size of scsi info buffer */
#define AFC_SCSI_INFO		256

/* size of memory mapped area */
#define AFC_MEM_SIZE		0x100000

/* size/number of outgoing message */
#define AFC_REPLY_SIZE		 16
#define AFC_REPLY_FRAMES	512

/* scsi parameters */
#define AFC_SCSI_ID		  7
#define AFC_SCSI_CAN_QUEUE	224
#define AFC_SCSI_CMD_PER_LUN	128
#define AFC_SCSI_SGLIST 	 32

#define AFC_SCSI_USE_CLUSTERING DISABLE_CLUSTERING

/* max retry messages */
#define OUTGOING_MAX_INIT	20

/* retry timer interval */
#define AFC_RETRY_PERIOD	2

/* stat timer interval */
#define AFC_STAT_PERIOD		4

/* ioctl timeout */
#define AFC_IOCTL_TIMEOUT	120*HZ
#define AFC_IOCTL_RETRY_MAX	10

/* ioctl buffer size (max:128K) */
#define MAX_IOCTL_BUFF		PAGE_SIZE

/* Debug Flag */
#undef AFC_DEBUG_PROC
#undef AFC_DEBUG_CHECK_CONDITION
#undef AFC_DEBUG_IRQ

/*******************************************************************************
 *	Supported hardware
 *		TOSHIBA RAID controller Series
 ******************************************************************************/
/* Generation Type 2 Device */
#define ER34XC_DEVICE_ID	0x0108
#define ER340C_SUBDEV_ID	0x0001
#define ER340C_CHANNELS		1
#define ER340C_CAN_QUEUE	224
#define ER340C_CMD_PER_LUN	127

/* Generation Type 3 Device */
#define ER341C_SUBDEV_ID	0x0002
#define ER341C_CHANNELS		1
#define ER341C_CAN_QUEUE	224
#define ER341C_CMD_PER_LUN	127

#define HRA38X_DEVICE_ID	0x3310
#define HRA380_SUBDEV_ID	0x3030
#define HRA380_CHANNELS		1
#define HRA380_CAN_QUEUE	224
#define HRA380_CMD_PER_LUN	127

#define AF38XC_DEVICE_ID	0x0109
#define T380_SUBDEV_ID		0x0001
#define T380_CHANNELS		1
#define T380_CAN_QUEUE		224
#define T380_CMD_PER_LUN	127

#define T340_SUBDEV_ID		0x0002
#define T340_CHANNELS		1
#define T340_CAN_QUEUE		224
#define T340_CMD_PER_LUN	127

#define AF380C_SUBDEV_ID	0x0003
#define AF380C_CHANNELS		1
#define AF380C_CAN_QUEUE	224
#define AF380C_CMD_PER_LUN	127

struct afc_boards {
	u8	name[8];	/* Board ID String */
	u8	generation;	/* Controller Generation Type */
	u8	cdb16_support;	/* Suuport 16 Byte CDB */
	u16	numPorts;	/* Number of SCSI ports */
	u16	can_queue;	/* Queue depth */
	u16	cmd_par_lun;	/* Command par lun */
};

/* AFC Proc Entry */
#define PROC_NAME_AFC_SCSI	"afxxxc"
#define PROC_LENGTH_AFC_SCSI	6

const char *afc_scsi_info(struct Scsi_Host *);
int afc_scsi_command(struct scsi_cmnd *);
int afc_scsi_queuecommand(struct scsi_cmnd *, void (*done)(struct scsi_cmnd *));
int afc_scsi_abort(struct scsi_cmnd *);
int afc_scsi_reset(struct scsi_cmnd *);

int afc_scsi_bios_param(struct scsi_device *,
	struct block_device *,sector_t, int *);

static int afc_scsi_proc_info(struct Scsi_Host *,
	char *, char **, off_t, int, int);

static int afc_probe_one(struct pci_dev *pdev, const struct pci_device_id *id);
static void afc_remove_one(struct pci_dev *);
#if LINUX_VERSION_CODE < LinuxVersionCode(2,6,15)
static void afc_shutdown(struct device *);
#else
static void afc_shutdown(struct pci_dev *);
#endif
static int afc_init(void);
static void afc_exit(void);


/* AFC SCSI Structures */
#define MAX_TARGETS		16
#define MAX_LUNS		 1
#define MAX_COMMANDS		256

/* AFC SCSI Message */
#define MSG_CDB_EXECUTE			0x0
#define MSG_MODE_CHANGE			0x1
#define MSG_SFR_RESET			0x2
#define MSG_BIOS_DATA_READ		0x3
#define MSG_BIOS_DATA_WRITE		0x4
#define MSG_POLLING_MODE_RESTART	0x5
#define MSG_LOGICAL_MODE_RESTART	0x6
#define MSG_OUTBOUND_MSG_FLUSH		0x7
#define MSG_RST_O_FREE_QUE	 	0x8
#define MSG_BUS_RESET			0xB
#define MSG_BUS_DEV_RESET		0xC
#define MSG_ABORT			0xD
#define MSG_CLEAR_QUEUE 		0xE

/* AFC SCSI Status */
#define STS_GOOD			0x00
#define STS_CHECK_CONDITION		0x02
#define STS_BUSY			0x08
#define STS_INTERMEDIATE_GOOD		0x10
#define STS_SELECTION_TIMEOUT		0x11
#define STS_DATA_OVERRUN_UNDERRUN	0x12
#define STS_UNEXPECTED_BUS_FREE		0x13
#define STS_INTERMEDIATE_C_GOOD		0x14
#define STS_RESERVATION_CONFLICT	0x18
#define STS_QUEUE_FULL			0x28
#define STS_MSG_COMPLETE		0xf0
#define STS_REQUEST_ABORTED		0xf1
#define STS_INVALID_HOST_ADAPTER	0xfd
#define STS_INVALID_MODE		0xfe
#define STS_MESSAGE_REJECT		0xff

/* AFC Sense Qualifier */
#define SENSEQUAL_RESET			0x92

/* AFC SCSI Command state */
#define SCMD_IDLE	0
#define SCMD_POSTED	1
#define SCMD_PENDING	2
#define SCMD_RETRIED	3

/* AFC SCSI Command */
struct afc_scmd
{
	struct list_head	list;
	struct scsi_cmnd	*cmd;
	u8			tag_no;
	u8			state;
	u16			message_type;
} __attribute__ ((packed));

/* AFC Target Information */
struct afc_target_info
{
	int outgoing_max;
	int outgoing_cnt;
} __attribute__ ((packed));

/* AFC Cache Mode Controls (s_Page3Cpara.mode) */
#define PAGE3C_SHUTDOWN			0x01
#define PAGE3C_PNP_START_ADAPTER	0x0E
#define PAGE3C_PNP_STOP_ADAPTER		0x0F
#define PAGE3C_CACHE_ON			0x10
#define PAGE3C_CACHE_OFF		0x11

/* Mode Parameter Page3C Table */
struct s_Page3Cpara
{
	u8	PageCode;		/*  00h : Page code = 0x3C */
	u8	ParaLength; 		/*  01h : Parameter length */
	u8	Mode;			/*  02h : Mode */
	union {
		 u8	LUNno;
		 u8	HDDno;
	} Target;			/*  03h : Target */
	u8	sts;			/*  04h : Status */
	u8	crfu[11];		/*  05h : rfu */
					/*  10h : Total Size */
} __attribute__ ((packed));

/* AFC Mode Parameter */
struct afc_parameter
{
	 struct ccs_modesel_head head;
	 union {
 		 struct s_Page3Cpara	page3c;
 		 u8			data[8];	/* dummy */
	 } para;
} __attribute__ ((packed));


/* AFC SCSI Host */

/* Resource data for each PCI AFC controller */
struct afc_pci
{
	struct pci_dev	*dev;
	u16		did;
	u16		sdid;
	int		irq;
} __attribute__ ((packed));

struct afc_lun_stat
{
	unsigned long	r_io_count;
	unsigned long	r_io_size;
	unsigned long	w_io_count;
	unsigned long	w_io_size;
} __attribute__ ((packed));

struct afc_scsi_host
{
	/* physical resources */
	volatile u32		*post_port;
	volatile u32		*reply_port;
	volatile u32		*irq_mask;
	unsigned long		mem_offset;
	unsigned long		mem_phys;
	unsigned long		csr_offset;
	unsigned long		csr_phys;
	void			*page_frame;
	dma_addr_t		page_frame_bus;
	struct afc_pci		pci;
	int			devnum;
	/* logical resources */
	struct Scsi_Host	*host;
	char			name[8];
	unsigned short		host_no;
	/* scmd control */
	struct list_head	free_list;
	struct list_head	exec_list[MAX_AFC_CHANNELS + 1];
	u32			reserved;
	/* flow control */
	u32			pending_flag;
	u32			reset_flag;
	struct afc_target_info	target_info[MAX_TARGETS];
	/* halt code */
	int			haltcode;
	/* statistics */
	unsigned long		queue_depth;
	unsigned long		max_queue_depth;
	unsigned long		scsi_cmds;
	unsigned long		post_msgs, reply_msgs;
	unsigned long		queue_full_sts, busy_sts;
	struct afc_lun_stat	persec_stat [MAX_TARGETS];
	struct afc_lun_stat	current_stat[MAX_TARGETS];
	/* buffer of scmds */
	struct afc_scmd		scmd[MAX_COMMANDS];
	/* buffer for configuration */
	struct scsi_device	control_device;
	struct scsi_cmnd	control_cmnd;
	struct afc_parameter	parameter_buf;
	struct request		control_req;
} __attribute__ ((packed));

/*	AFC SCSI sg list */
struct afc_scsi_sglist
{
	u32 address_40:8,
	    count:24;
	u32 address;
} __attribute__ ((packed));


/* AFC SCSI post/reply message */
#define AFC_MAX_SGLISTS 11
struct scsi_post_message
{
	u64	scmd_ptr;	/* struct afc_scmd *scmd; */
	u8	dma_count;
	u8	message;
	u8	number;
	u8	length;
	u8	queue_tag;
	u8	queue_code;
	u8	lun;
	u8	target_id:4,
		bus_no:4;
	union
	{
		struct
		{
			u8	cdb[12];
			u8	cdb_size;
			u8	busy_status;
			u8	check_condition;
			u8	ref_count;
			/* dma request follows */
			struct afc_scsi_sglist	sglist[AFC_MAX_SGLISTS];
			u32	reserved;
		}cdb12;
		struct
		{
			u8	cdb[16];
			/* dma request follows */
			struct afc_scsi_sglist	sglist[AFC_MAX_SGLISTS];
			u8	cdb_size;
			u8	busy_status;
			u8	check_condition;
			u8	ref_count;
		}cdb16;
	}u;
	u32	next_ptr;	/* struct scsi_post_message *next; */
} __attribute__ ((packed));

struct scsi_reply_message
{
	u64	scmd_ptr;	/* struct afc_scmd *scmd; */
	u8	qualifier;
	u8	code;
	u8	key;
	u8	status;
	u32	count;
} __attribute__ ((packed));

/* Messenger inlines */
extern inline u32 AFC_POST_READ32(struct afc_scsi_host *h)
{
	return readl(h->post_port);
}

extern inline void AFC_POST_WRITE32(struct afc_scsi_host *h, u32 val)
{
	writel(val, h->post_port);
}

extern inline u32 AFC_REPLY_READ32(struct afc_scsi_host *h)
{
	return readl(h->reply_port);
}

extern inline void AFC_REPLY_WRITE32(struct afc_scsi_host *h, u32 val)
{
	writel(val, h->reply_port);
}

extern inline u32 AFC_IRQ_READ32(struct afc_scsi_host *h)
{
	return readl(h->irq_mask);
}

extern inline void AFC_IRQ_WRITE32(struct afc_scsi_host *h, u32 val)
{
	writel(val, h->irq_mask);
}

extern inline void afc_post_message(struct afc_scsi_host *h, u32 m)
{
	/* The second line isnt spurious - thats forcing PCI posting */
	AFC_POST_WRITE32(h,m);
	(void) AFC_IRQ_READ32(h);
	mb();
}

extern inline void afc_flush_reply(struct afc_scsi_host *h, u32 m)
{
	AFC_REPLY_WRITE32(h,m);
}

/* AFC BIOS Access Mode & Auto Sense Mode */
#define DISABLE_BIOS_ACCESS		0x80000000
#define LINUX_MODE_ENABLE		0x02000000

/* Generation Type 2 Registers  */
#define IOP33X_ATU_OMWTVR0		0x60
#define IOP33X_ATU_OUMWTVR0		0x64
#define IOP33X_BIOS_COMMAND_REG		0x58
#define IOP33X_HALTCODE_REG		0xBC

/* Generation Type 3 Registers */
#define IOP348_ATU_OMWTVR_L		0xD0000
#define IOP348_ATU_OMWTVR_H		0xD0004
#define IOP348_BIOS_COMMAND_REG		0x5C
#define IOP348_HALTCODE_REG		0x94


#endif

#endif
