#ifndef _kaho_def_h
#define _kaho_def_h

/* ------------------------------------------------------------------
 * Definitions for x86_64 & i386
 * ---------------------------------------------------------------- */
#if defined(__x86_64__) || defined(__i386__)

#ifdef CONFIG_KAHO_RT
static const unsigned char trap_code = 0xcc;
static const int trap_code_size = 1;
#endif

#endif

/* ------------------------------------------------------------------
 * Definitions for x86_64
 * ---------------------------------------------------------------- */
#if defined(__x86_64__)

static const unsigned char jmp_code_abs[] = {
	/* pushq $0x12345678 */
	0x68, 0x12, 0x34, 0x56, 0x78,
	/* movl  $0x9abcdef0, 4(%rsp) */
	0xc7, 0x44, 0x24, 0x04, 0x9a, 0xbc, 0xde, 0xf0,
	/* retq */
	0xc3
};

static const unsigned char jmp_code_rel[] = {
	/* jmpq  $0x12345678 */
	0xe9, 0x12, 0x34, 0x56, 0x78,
};

/* size of jump code */
static const int jmp_code_size_abs = sizeof(jmp_code_abs);
static const int jmp_code_size_rel = sizeof(jmp_code_rel);
#define ARCH_JMP_CODE_SIZE	sizeof(jmp_code_abs)

inline void arch_set_pc(struct pt_regs* regs, unsigned long pc)
{
	regs->rip = pc;
}

inline unsigned long arch_get_pc(struct pt_regs* regs)
{
	return regs->rip;
}

#endif

/* ------------------------------------------------------------------
 * Definitions for i386
 * ---------------------------------------------------------------- */
#if defined(__i386__)
static const unsigned char jmp_code_rel[] = {
	/* jmpq  $0x12345678 */
	0xe9, 0x12, 0x34, 0x56, 0x78,
};
static const int jmp_code_size_rel = sizeof(jmp_code_rel);
#define ARCH_JMP_CODE_SIZE sizeof(jmp_code_rel)

inline void arch_set_pc(struct pt_regs* regs, unsigned long pc)
{
	regs->eip = pc;
}

inline unsigned long arch_get_pc(struct pt_regs* regs)
{
	return regs->eip;
}
#endif

/* ----------------------------------------------------------------
 *
 * Definictions for all architecutres
 *
 * ---------------------------------------------------------------- */
#define HANDLE_NOT_USED	0
#define HANDLE_USED	1

#define NO_PATCHER	-1

#define MEM_ALLOC	0
#define MEM_FREE	1

#define REQ_ACT		0
#define REQ_DEACT	1
#define REQ_ACT_RT	2
#define REQ_DEACT_RT	3

#define NOT_CODE_BKUP	0
#define CODE_BKUP	1
#define CODE_RESTORE	2

#define CANCELED	1

#ifdef CONFIG_KAHO_RT
#define THR_NO_MARK	0
#define THR_MARKED	1
#define THR_ESRCH	2
#define THR_NEWFACE	3
#define THR_CHK_ERR	-1

#define MAX_RETRY_RT_ACT	10
#endif

struct kaho_patch_item {
	int patch_type;
	unsigned long patch_addr;  /* address to patch in the target */
	unsigned long patch_size;  /* size of the patch */
	unsigned long target_addr; /* address to be patched */
	unsigned char backup[ARCH_JMP_CODE_SIZE];
#ifdef CONFIG_KAHO_RT
	atomic_t backup_done;
#endif
	int flags;
};

struct kaho_mem_alloc_struct {
	/* for memory allocation */
	unsigned long req_addr;
	unsigned long req_size;
	unsigned long act_addr;
};

#ifdef CONFIG_KAHO_RT
struct kaho_patch_info;
struct kaho_thr_list {
	struct utrace_attached_engine* engine;
	struct rcu_head rcu_head;
	struct list_head chk_list; /* list of threads to be checked */
	struct list_head thr_list; /* list of all threads in the process */
	struct task_struct* task;
	struct kaho_patch_info* kpi;
	atomic_t checked;
	atomic_t used;
	int sem_upped;
};
#endif

struct kaho_patch_info {
	/* basic variables */
	struct list_head list;

	/* lock for the members of this structure */
	struct semaphore sem;
	atomic_t used; /* reference counter */

	int handle;
	atomic_t state;
	struct task_struct* target_task;
	struct utrace_attached_engine* engine;

	/* patch item array */
	struct kaho_mem_alloc_struct* kmas;
	struct semaphore mem_ope_sem;
	int nr_sub_patch;
	struct kaho_patch_item* items;
	int canceled;
	u32 exec_id;

	/* for delayed release */
	struct timer_list timer;
	unsigned long nvcsw, nivcsw;

#ifdef CONFIG_KAHO_RT
	/* list of kaho_patch_info that is waiting for realtime activation */
	struct list_head trap_list;	/* it is linked from ... */
	struct list_head chk_list;	/* List head for the */
	struct list_head thr_list;	/* List head for the */
	/*spinlock_t chk_list_lock;
	spinlock_t thr_list_lock;*/
	spinlock_t list_lock;
	int jmp_code_size;
	struct semaphore rt_sem;

	/* list of thread information */
#endif
};

#endif

