/*
 *  mazedev.h, Common data and definitions for MAZE monitoring functions.
 *
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */

#ifndef _LINUX_MAZE_H
#define _LINUX_MAZE_H

#include <linux/ioctl.h>
#include <linux/sched.h>
#include <asm/cputime.h>

#define MZDEV_NAME          "maze"
#define MZCLASS_NAME        MZDEV_NAME

#define MAZE_IOC_MAGIC		0xf0

#define MAZE_IOCSETINFO		_IOW(MAZE_IOC_MAGIC, 1, int *)
#define MAZE_IOCGETINFO		_IOR(MAZE_IOC_MAGIC, 2, int *)
#define MAZE_IOCGETNOTICE	_IOR(MAZE_IOC_MAGIC, 3, int *)
#define MAZE_IOCSETATTR		_IOW(MAZE_IOC_MAGIC, 4, int *)

#define MAZE_IOC_MAXNR		4

#define MAZE_INHRCHILD		0x00000001
#define MAZE_INHRTHREAD		0x00000002

#define MAZE_ATTR_INHRCHILD	0x00000001
#define MAZE_ATTR_INHRTHREAD	0x00000002
#define MAZE_ATTR_MONITORALL	0x00000004

#define MAZE_LIM_USEDEFAULT	(-1)
#define MAZE_LIM_NONE        0
#define MAZE_SIG_USEDEFAULT	(-1)
#define MAZE_SIG_NONE        0

#define MAZE_LIM_MAX        655350

#define MAZE_LIM_SOFTDEF	0
#define MAZE_LIM_HARDDEF	0
#define MAZE_SIG_SOFTDEF	SIGXCPU
#define MAZE_SIG_HARDDEF	SIGKILL

#define MAZE_EV_SLIMOVER	0
#define MAZE_EV_HLIMOVER	1
#define MAZE_EV_EXITTARGET	2


struct maze_data {
	int pid;
	int soft_lim;
	int hard_lim;
	int soft_sig;
	int hard_sig;
	int time;
	int flag;
};

struct maze_notice {
	int pid;
	int event;
	struct timeval tv;
	int data1;
	int data2;
};


#ifdef __KERNEL__

#define MZDEV_MINOR             0
#define MZDEV_NUM               1

#ifdef CONFIG_MAZEDEBUG
#define DBGPRINT(str, ...) printk("[%s] " str, __FUNCTION__, ##__VA_ARGS__);
#define __DBGPRINT(str, ...) printk(str, ##__VA_ARGS__);
#else
#define DBGPRINT(str, ...)
#define __DBGPRINT(str, ...)
#endif

#define ERRPRINT(str, ...) printk("ERROR [%s] " str, __FUNCTION__, ##__VA_ARGS__);

/*
 * Maze hash table data
 */
#define MZHASH_SHIFT    12
#define MZHASH_NR       (1<<MZHASH_SHIFT)
#define mz_hashfn(nr) hash_long((unsigned long)nr, MZHASH_SHIFT)

/*
 * Sent signal flag.
 */
#define NOT_SEND_SIG 0
#define SENT_SOFT_SIG 1
#define SENT_HARD_SIG 2

struct maze_entry {
	struct hlist_node chain;
	int pid;
	int soft_lim;
	int hard_lim;
	int soft_sig;
	int hard_sig;
	int sig_flag;
	cputime_t reset_utime;
	cputime_t reset_stime;
	int flag;
    unsigned long nvcsw;
    unsigned long nivcsw;
	struct rcu_head rcu_head;
};

/*
 * Notice data
 */
struct notice_entry {
	struct list_head chain;
	struct maze_notice mn;
};

static inline cputime_t get_maze_count(struct maze_entry *entry, struct task_struct *task)
{
	return (task->utime - entry->reset_utime) + (task->stime - entry->reset_stime);
}

#endif /* __KERNEL__ */

#endif /* _LINUX_MAZE_H */
