/*
 *  mazequeue.h
 *
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */

#ifndef _LINUX_MAZEQUE_H
#define _LINUX_MAZEQUE_H

#ifdef __KERNEL__

#include <linux/slab.h>
#include <linux/spinlock.h>
#include "maze.h"

#define ENQUEUE_MAZE 1
#define DEQUEUE_MAZE (1 << 1)
#define REBALANCE_MAZE (DEQUEUE_MAZE | ENQUEUE_MAZE)
#define MIGRATION_MAZE ((1 << 2) | REBALANCE_MAZE)
#define INTO_SLEEPING_MAZE ((1 << 3) | MIGRATION_MAZE)
#define WAKEUP_MAZE ((1 << 4) | MIGRATION_MAZE)
#define NOTHING_MAZE (WAKEUP_MAZE | INTO_SLEEPING_MAZE)
#define DEQUEUE_AND_FREE_MAZE (~0)

#define CONFIG_MAZE_HASH_QUEUE

#ifdef CONFIG_MAZE_RB_QEUEUE
#include "mazequeue_rb.h"
#elif defined(CONFIG_MAZE_HASH_QUEUE)
#include "mazequeue_hash.h"
#else
#include "mazequeue_list.h"
#endif

#define maze_queue_entry(ptr) \
	container_of(ptr, struct maze_queue_entry, entry)

#define maze_entry(ptr) (&ptr->entry)

/*
 * 
 * Note. Must be called rcu_read_lock().
 */
#define for_each_queue(itr, qentry)  \
	                   for (__exec_cmds(),__init_queue_itr(itr),qentry = next_queue_entry(itr); \
						   qentry;					\
						   qentry = next_queue_entry(itr))

struct maze_queue_ops
{
	int (*compare)(struct maze_queue_entry *, struct maze_queue_entry *);
};

static inline struct maze_queue_entry* alloc_queue_entry(void)
{
	return kzalloc(sizeof(struct maze_queue_entry), GFP_ATOMIC);
}

static inline void free_queue_entry(struct maze_queue_entry *qentry)
{
	kfree(qentry);
}

extern void init_queue(struct maze_queue_ops *ops);

extern void enqueue(struct maze_queue_entry *qentry);

extern void migration(struct maze_queue_entry *qentry);

extern void dequeue(struct maze_queue_entry *qentry);

extern void dequeue_and_free(struct maze_queue_entry *qentry);

extern void rebalance(struct maze_queue_entry *qentry);

extern void might_sleep_list(struct maze_queue_entry *qentry, struct task_struct *task);

extern void __exec_cmds(void);

extern void lock_all_queue(void);

extern void unlock_all_queue(void);

extern int get_amount_queue(int cpu);

extern int get_sleeping_list_len(int cpu);

extern void fastcall __init_queue_itr(struct maze_queue_itr *itr);

extern struct maze_queue_entry* fastcall next_queue_entry(struct maze_queue_itr *itr);

extern int procmz_queue_debug_seq_show(struct seq_file *s, void *v);

#endif /* __KERNEL__ */
#endif /* _LINUX_MAZEQUE_H */
