/*
 *  mazeque_list.c
 *
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */
#include <linux/cpumask.h>
#include <linux/list.h>
#include <linux/percpu.h>
#include <linux/slab.h>
#include <linux/seq_file.h>
#include "maze.h"
#include "mazequeue.h"

struct maze_queue
{
	struct list_head head;
	spinlock_t lock;
	struct maze_queue_ops *ops;
};

DEFINE_PER_CPU(struct maze_queue, maze_queue);
DEFINE_PER_CPU(unsigned long, maze_queue_lock_flags);

static inline void __init_queue(int cpu, struct maze_queue_ops *not_use)
{
	INIT_LIST_HEAD(&per_cpu(maze_queue, cpu).head);
	spin_lock_init(&per_cpu(maze_queue, cpu).lock);
}

static inline void insert_entry_on(struct maze_queue_entry *qentry, int cpu)
{
	list_add_tail(&qentry->qlist, &per_cpu(maze_queue, cpu).head);
	qentry->state |= ENQUEUE_MAZE;
}

static inline void remove_entry_on(struct maze_queue_entry *qentry, int not_use)
{
	if (!(qentry->state & ENQUEUE_MAZE))
		return;
	list_del(&qentry->qlist);
}

void fastcall __init_queue_itr(struct maze_queue_itr *itr)
{
	itr->queue = &__get_cpu_var(maze_queue);
	itr->curr = itr->queue->head.next;
}

struct maze_queue_entry* fastcall next_queue_entry(struct maze_queue_itr *itr)
{
	struct maze_queue_entry *qentry;

	if (itr->curr == &itr->queue->head)
		return NULL;

	qentry = list_entry(itr->curr, struct maze_queue_entry, qlist);
	itr->curr = itr->curr->next;

	return qentry;
}

#ifdef CONFIG_MAZEPROCDEBUG
int procmz_queue_debug_seq_show(struct seq_file *s, void *v)
{
	return 0;
}
#endif

#include "queue_internal.c"

