/*
 *  mazeque_rb.c
 *
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */
#include <linux/cpumask.h>
#include <linux/list.h>
#include <linux/percpu.h>
#include <linux/rbtree.h>
#include "maze.h"
#include "mazequeue.h"

struct maze_queue
{
	struct rb_root root;
	struct maze_queue_entry *most_left;
	spinlock_t lock;
	struct maze_queue_ops *ops;
};

DEFINE_PER_CPU(struct maze_queue, maze_queue);
DEFINE_PER_CPU(unsigned long, maze_queue_lock_flags);

static inline void __init_queue(int cpu, struct maze_queue_ops *ops)
{
	per_cpu(maze_queue, cpu).root.rb_node = NULL;
	per_cpu(maze_queue, cpu).most_left = NULL;
	per_cpu(maze_queue, cpu).ops = ops;
	spin_lock_init(&per_cpu(maze_queue, cpu).lock);
}

static void __insert_entry_on(struct maze_queue_entry *qentry, int cpu)
{
	struct rb_root *root = &per_cpu(maze_queue, cpu).root;
	struct rb_node **p = &root->rb_node;
	struct rb_node *parent = NULL;
	struct maze_queue_entry *curr;
	int r;

	while (*p) {
		parent = *p;
		curr = rb_entry(parent, struct maze_queue_entry, node);
		r = per_cpu(maze_queue, cpu).ops->compare(qentry, curr);

		if (r < 0)
			p = &(*p)->rb_left;
		else if (r >= 0)
			p = &(*p)->rb_right;
	}

	rb_link_node(&qentry->node, parent, p);
}

static inline void insert_entry_on(struct maze_queue_entry *qentry, int cpu)
{
	__insert_entry_on(qentry, cpu);
	rb_insert_color(&qentry->node, &per_cpu(maze_queue, cpu).root);

	if (per_cpu(maze_queue, cpu).most_left)
		per_cpu(maze_queue, cpu).most_left =
			rb_entry(rb_first(&per_cpu(maze_queue, cpu).root), struct maze_queue_entry, node);
	else
		per_cpu(maze_queue, cpu).most_left = qentry;

	qentry->state |= ENQUEUE_MAZE;
}

static inline void remove_entry_on(struct maze_queue_entry *qentry, int cpu)
{
	struct rb_node *n;

	if (!(qentry->state & ENQUEUE_MAZE))
		return;

	if (per_cpu(maze_queue, cpu).most_left == qentry) {
		n = rb_next(&qentry->node);
		if (n) {
			per_cpu(maze_queue, cpu).most_left =
				rb_entry(n, struct maze_queue_entry, node);
		} else {
			per_cpu(maze_queue, cpu).most_left = NULL;
		}
	}
	rb_erase(&qentry->node, &per_cpu(maze_queue, cpu).root);
	qentry->node.rb_parent_color = 0;
	qentry->node.rb_right = NULL;
	qentry->node.rb_left = NULL;
}

void fastcall __init_queue_itr(struct maze_queue_itr *itr)
{
	itr->curr = __get_cpu_var(maze_queue).most_left;
}

struct maze_queue_entry* fastcall next_queue_entry(struct maze_queue_itr *itr)
{
	struct rb_node *next;
	struct maze_queue_entry *qentry;

	if (!itr->curr) {
		return NULL;
	}

	qentry = rb_entry(&itr->curr->node, struct maze_queue_entry, node);
	next = rb_next(&itr->curr->node);
	if (next) {
		itr->curr = rb_entry(next, struct maze_queue_entry, node);
	} else {
		itr->curr = NULL;
	}
	return qentry;
}

#ifdef CONFIG_MAZEPROCDEBUG
int procmz_queue_debug_seq_show(struct seq_file *s, void *v)
{
	return 0;
}
#endif

#include "queue_internal.c"

