/*
 *  mazedev.h
 *
 *  Copyright (C) 2007-2008 MIRACLE LINUX Corp.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 */

#ifndef _LINUX_MAZE_TIMER_H
#define _LINUX_MAZE_TIMER_H

#ifdef __KERNEL__

#include <linux/cpumask.h>
#include <linux/smp.h>
#include <linux/timer.h>
#include <linux/percpu.h>
#include "maze.h"

#ifdef CONFIG_MAZEPROCDEBUG
#include <asm/tce.h>
#endif

DECLARE_PER_CPU(struct timer_list, maze_timer);
DECLARE_PER_CPU(atomic_t, started_maze_timer);


static inline void continue_timer(void (*function)(unsigned long), int interval)
{
	__get_cpu_var(maze_timer).function = function;
	__get_cpu_var(maze_timer).expires = jiffies + interval;
	init_timer(&__get_cpu_var(maze_timer));
	add_timer(&__get_cpu_var(maze_timer));
}

/*
 *  Set timer to check maze entry
 */
static inline void set_cpu_timer(void (*function)(unsigned long), int interval)
{
	preempt_disable();
	if (unlikely(!atomic_xchg(&__get_cpu_var(started_maze_timer), 1))){
		DBGPRINT("start timer for CPU = %d\n", smp_processor_id());
		continue_timer(function, interval);
	}
       preempt_enable();
}

/*
 *  Delete all cpu timer to check maze entry
 */
static inline void del_all_cpu_timer_sync(void)
{
	int i;

	for_each_possible_cpu(i) {
		if (atomic_xchg(&per_cpu(started_maze_timer, i), 1))
			del_timer_sync(&per_cpu(maze_timer, i));
	}
}

static inline int is_active_timer(int cpu)
{
	return atomic_read(&per_cpu(started_maze_timer, cpu));
}


#ifdef CONFIG_MAZEPROCDEBUG
DECLARE_PER_CPU(unsigned long, peek_time);
DECLARE_PER_CPU(unsigned long, _int_time);
DECLARE_PER_CPU(unsigned long, _timer_count);

static inline unsigned long long begin_timer(void)
{
	return get_cycles();
}

static inline void after_timer(unsigned long long igl)
{
	unsigned long long t = get_cycles() - igl;

	if (__get_cpu_var(peek_time) < t)
		__get_cpu_var(peek_time) = t;

	__get_cpu_var(_timer_count)++;
	__get_cpu_var(_int_time) += t;
}

static inline unsigned long long get_peek_time(int cpu)
{
	return per_cpu(peek_time, cpu);
}

static inline unsigned long long get_ave_time(int cpu)
{
	unsigned long long t;

	if (unlikely(!per_cpu(_timer_count, cpu)))
		return 0;

	t = per_cpu(_int_time, cpu) / per_cpu(_timer_count, cpu);

	per_cpu(_timer_count, cpu) = 0;
	per_cpu(_int_time, cpu) = 0;

	return t;
}

#else
static inline unsigned long long begin_timer(void)
{
	return 0;
}

static inline void after_timer(unsigned long long igl)
{
}

static inline unsigned long long get_peek_time(int cpu)
{
	return 0;
}

static inline unsigned long long get_ave_time(int cpu)
{
	return 0;
}

#endif

#endif /* __KERNEL__  */
#endif /* _LINUX_MAZE_TIMER_H */
